@@ -1,6 +1,7 @@
 """Base package for gate API code."""
 import abc
 import base64
+from datetime import datetime, timedelta
 from hashlib import sha1
 import json
 import secrets
@@ -11,13 +12,15 @@ from xml.etree.ElementTree import Element  # nosec
 from Crypto.Cipher import AES  # nosec
 from Crypto.Cipher._mode_cbc import CbcMode  # nosec
 from defusedxml import ElementTree
-import requests
-from requests import Response
+from httpx import AsyncClient
 from typing_extensions import Final
 
 from .common import (
-    AbstractDoor,
+    CLOSE_DOOR_STATUSES,
+    OPEN_DOOR_STATUSES,
+    AllDoorStatus,
     ApiError,
+    CachedTransitionDoorStatus,
     CredentialsIncorrectException,
     CredentialsNotSetException,
     DoorNotSetException,
@@ -31,12 +34,15 @@ from .common import (
     ISmartGateDoor,
     ISmartGateInfoResponse,
     RequestOption,
+    TransitionDoorStatus,
     element_to_api_error,
     element_to_gogogate2_activate_response,
     element_to_gogogate2_info_response,
     element_to_ismartgate_activate_response,
     element_to_ismartgate_info_response,
+    get_configured_doors,
     get_door_by_id,
+    InvalidDoorException,
 )
 from .const import GogoGate2ApiErrorCode, ISmartGateApiErrorCode
 
@@ -46,28 +52,26 @@ class ApiCipher:
 
     def __init__(self, key: str) -> None:
         """Initialize the object."""
-        self._key: Final[str] = key
-        self._key_bytes: Final[bytes] = key.encode("utf-8")
+        self._key: Final = key
+        self._key_bytes: Final = key.encode("utf-8")
 
     def encrypt(self, content: str, init_vector: Optional[str] = None) -> str:
         """Encrypt content."""
-        init_vector_bytes: Final[bytes] = ApiCipher.pad_pkcs5(
+        init_vector_bytes: Final = ApiCipher.pad_pkcs5(
             init_vector or uuid.uuid4().hex
         ).encode("utf-8")[: AES.block_size]
-        content_bytes: Final[bytes] = ISmartGateApiCipher.pad_pkcs5(content).encode(
-            "utf-8"
-        )
-        cipher: Final[CbcMode] = cast(
+        content_bytes: Final = ISmartGateApiCipher.pad_pkcs5(content).encode("utf-8")
+        cipher: Final = cast(
             CbcMode, AES.new(self._key_bytes, AES.MODE_CBC, init_vector_bytes)
         )
-        encrypted_bytes: Final[bytes] = cipher.encrypt(content_bytes)
+        encrypted_bytes: Final = cipher.encrypt(content_bytes)
         return str(init_vector_bytes + base64.b64encode(encrypted_bytes), "utf-8")
 
     def decrypt(self, content: str) -> str:
         """Decrypt content."""
-        init_vector: Final[bytes] = content.encode("utf-8")[: AES.block_size]
-        encrypted_bytes: Final[bytes] = base64.b64decode(content[AES.block_size :])
-        cipher: Final[CbcMode] = cast(
+        init_vector: Final = content.encode("utf-8")[: AES.block_size]
+        encrypted_bytes: Final = base64.b64decode(content[AES.block_size :])
+        cipher: Final = cast(
             CbcMode, AES.new(self._key_bytes, AES.MODE_CBC, init_vector)
         )
         return ApiCipher.unpad_pkcs5(cipher.decrypt(encrypted_bytes)).decode("utf-8")
@@ -75,7 +79,7 @@ class ApiCipher:
     @staticmethod
     def pad_pkcs5(data: str) -> str:
         """Add padding to bytes."""
-        block_size: Final[int] = AES.block_size
+        block_size: Final = AES.block_size
         return data + (block_size - len(data) % block_size) * chr(
             block_size - len(data) % block_size
         )
@@ -89,7 +93,7 @@ class ApiCipher:
 class GogoGate2ApiCipher(ApiCipher):
     """Cipher for GogoGate2 devices."""
 
-    SHARED_SECRET: Final[str] = "0e3b7%i1X9@54cAf"
+    SHARED_SECRET: Final = "0e3b7%i1X9@54cAf"
 
     def __init__(self) -> None:
         """Initialize the object."""
@@ -99,21 +103,21 @@ class GogoGate2ApiCipher(ApiCipher):
 class ISmartGateApiCipher(ApiCipher):
     """Cipher for iSmartGate devices."""
 
-    RAW_TOKEN_FORMAT: Final[str] = "%s@ismartgate"
+    RAW_TOKEN_FORMAT: Final = "%s@ismartgate"
 
     def __init__(self, username: str, password: str) -> None:
         """Initialize the object."""
-        self._username: Final[str] = username
-        self._password: Final[str] = password
+        self._username: Final = username
+        self._password: Final = password
 
         # Calculate the token.
         raw_token: Final[
             str
         ] = ISmartGateApiCipher.RAW_TOKEN_FORMAT % self._username.lower()
-        self._token: Final[str] = sha1(raw_token.encode("utf-8")).hexdigest()  # nosec
+        self._token: Final = sha1(raw_token.encode("utf-8")).hexdigest()  # nosec
 
         # Calculate the key and pass it onto the superclass.
-        sha1_hex_str: Final[str] = sha1(  # nosec
+        sha1_hex_str: Final = sha1(  # nosec
             (self._username.lower() + self._password).encode()
         ).hexdigest()
 
@@ -139,22 +143,34 @@ ActivateResponseType = TypeVar(
 )
 
 
+# pylint: disable=too-many-instance-attributes
 class AbstractGateApi(
     Generic[ApiCipherType, InfoResponseType, ActivateResponseType], abc.ABC
 ):
     """API capable of communicating with a gogogate2 devices."""
 
-    API_URL_TEMPLATE: Final[str] = "http://%s/api.php"
+    API_URL_TEMPLATE: Final = "http://%s/api.php"
+    DEFAULT_REQUEST_TIMEOUT = timedelta(seconds=20)
+    DEFAULT_TRANSITION_STATUS_TIMEOUT = timedelta(seconds=55)
 
     def __init__(
-        self, host: str, username: str, password: str, api_cipher: ApiCipherType
+        self,
+        host: str,
+        username: str,
+        password: str,
+        api_cipher: ApiCipherType,
+        request_timeout: timedelta = DEFAULT_REQUEST_TIMEOUT,
+        transition_status_timeout: timedelta = DEFAULT_TRANSITION_STATUS_TIMEOUT,
     ) -> None:
         """Initialize the object."""
-        self._host: Final[str] = host
-        self._username: Final[str] = username
-        self._password: Final[str] = password
-        self._cipher: Final[ApiCipherType] = api_cipher
-        self._api_url: Final[str] = AbstractGateApi.API_URL_TEMPLATE % host
+        self._host: Final = host
+        self._username: Final = username
+        self._password: Final = password
+        self._cipher: Final = api_cipher
+        self._request_timeout: Final = request_timeout
+        self._transition_status_timeout: Final = transition_status_timeout
+        self._api_url: Final = AbstractGateApi.API_URL_TEMPLATE % host
+        self._transition_door_status: Final[Dict[int, CachedTransitionDoorStatus]] = {}
 
     @property
     def host(self) -> str:
@@ -176,13 +192,13 @@ class AbstractGateApi(
         """Get the cipher."""
         return self._cipher
 
-    def _request(
+    async def _async_request(
         self,
         option: RequestOption,
         arg1: Optional[str] = None,
         arg2: Optional[str] = None,
     ) -> Element:
-        command_str: Final[str] = json.dumps(
+        command_str: Final = json.dumps(
             (
                 self._username,
                 self._password,
@@ -192,40 +208,41 @@ class AbstractGateApi(
             )
         )
 
-        response: Final[Response] = requests.get(
-            self._api_url,
-            params={
-                "data": self._cipher.encrypt(command_str),
-                **self._get_extra_url_params(),
-            },
-            timeout=5,
-        )
-        response_raw: Final[str] = response.content.decode("utf-8")
+        async with AsyncClient() as client:
+            response: Final = await client.get(
+                self._api_url,
+                params={
+                    "data": self._cipher.encrypt(command_str),
+                    **self._get_extra_url_params(),
+                },
+                timeout=self._request_timeout.seconds,
+            )
+        response_raw: Final = response.content.decode("utf-8")
 
         try:
             # Error messages are returned unencrypted so we try to decrypt the response.
             # if that fails, then we use what was returned.
-            response_text = self._cipher.decrypt(response.content.decode("utf-8"))
+            response_text = self._cipher.decrypt(response_raw)
         except ValueError:
             response_text = response_raw
 
-        root_element: Final[Element] = ElementTree.fromstring(response_text)
+        root_element: Final = ElementTree.fromstring(response_text)
 
-        error_element: Final[Optional[Element]] = root_element.find("error")
+        error_element: Final = root_element.find("error")
         if error_element:
-            api_error: Final[ApiError] = element_to_api_error(error_element)
+            api_error: Final = element_to_api_error(error_element)
             raise self._get_exception_map().get(api_error.code, ApiError)(
                 api_error.code, api_error.message
             )
 
-        return root_element
+        return cast(Element, root_element)
 
     @abc.abstractmethod
-    def info(self) -> InfoResponseType:
+    async def async_info(self) -> InfoResponseType:
         """Get info about the device and doors."""
 
     @abc.abstractmethod
-    def activate(self, door_id: int) -> ActivateResponseType:
+    async def async_activate(self, door_id: int) -> ActivateResponseType:
         """Send a command to open/close/stop the door.
 
         Devices do not have a status for opening or closing. So running
@@ -246,11 +263,11 @@ class AbstractGateApi(
     def _get_extra_url_params(self) -> Dict[str, str]:
         return {}
 
-    def _info(self) -> Element:
+    async def _async_info(self) -> Element:
         """Get info about the device and doors."""
-        return self._request(RequestOption.INFO)
+        return await self._async_request(RequestOption.INFO)
 
-    def _activate(
+    async def _async_activate(
         self, door_id: int, info: Optional[InfoResponseType] = None
     ) -> Element:
         """Send a command to open/close/stop the door.
@@ -259,75 +276,160 @@ class AbstractGateApi(
         running this method during an action will stop the door. It's
         recommended you use open_door() or close_door() as those methods check
         the status before running and run if needed."""
-        return self._request(
+        return await self._async_request(
             RequestOption.ACTIVATE,
             str(door_id),
-            self._get_activate_api_code(info if info else self.info(), door_id),
+            self._get_activate_api_code(
+                info if info else await self.async_info(), door_id
+            ),
         )
 
-    def _set_door_status(self, door_id: int, door_status: DoorStatus) -> bool:
+    async def _async_set_door_status(
+        self,
+        door_id: int,
+        target_door_status: DoorStatus,
+        consider_transitional_states: bool = True,
+    ) -> bool:
         """Send call to open/close a door if door is not already in that state."""
-        if door_status == DoorStatus.UNDEFINED:
+        if target_door_status == DoorStatus.UNDEFINED:
             return False
 
         # Get current door status.
-        info: Final[InfoResponseType] = self.info()
-        door: Final[Optional[AbstractDoor]] = get_door_by_id(door_id, info)
+        info: Final = await self.async_info()
+        statuses: Final = self._get_door_statuses(
+            info, use_transitional_status=consider_transitional_states
+        )
+        current_door_status: Final = statuses.get(door_id)
+        result_door_statuses: Final = OPEN_DOOR_STATUSES if target_door_status == DoorStatus.OPENED else CLOSE_DOOR_STATUSES
+        transitional_door_status: Final = TransitionDoorStatus.OPENING if target_door_status == DoorStatus.OPENED else TransitionDoorStatus.CLOSING
 
-        # No door found or door already at desired status.
-        if not door or door.status == door_status:
+        # Door is invalid, not configured, already in desired state or transitioning to it.
+        if not current_door_status or current_door_status in result_door_statuses:
             return False
 
-        self._activate(door_id, info)
+        await self._async_activate(door_id, info)
+
+        # Set the transitional status.
+        self._transition_door_status[door_id] = CachedTransitionDoorStatus(
+            door_id=door_id,
+            activated=datetime.utcnow(),
+            transition_status=transitional_door_status,
+            target_status=target_door_status,
+        )
+
         return True
 
-    def close_door(self, door_id: int) -> bool:
+    async def async_close_door(
+        self, door_id: int, consider_transitional_states: bool = True
+    ) -> bool:
         """Close a door.
 
         :return True if close command sent, False otherwise.
         """
-        return self._set_door_status(door_id, DoorStatus.CLOSED)
+        return await self._async_set_door_status(
+            door_id,
+            DoorStatus.CLOSED,
+            consider_transitional_states=consider_transitional_states,
+        )
 
-    def open_door(self, door_id: int) -> bool:
+    async def async_open_door(
+        self, door_id: int, consider_transitional_states: bool = True
+    ) -> bool:
         """Open a door.
 
         :return True if open command sent, False otherwise.
         """
-        return self._set_door_status(door_id, DoorStatus.OPENED)
+        return await self._async_set_door_status(
+            door_id,
+            DoorStatus.OPENED,
+            consider_transitional_states=consider_transitional_states,
+        )
+
+    def _get_door_statuses(
+        self, info: InfoResponseType, use_transitional_status: bool = True
+    ) -> Dict[int, AllDoorStatus]:
+        doors: Final = get_configured_doors(info)
+
+        # Clean out the cache.
+        for (cached_door_id, cached_transitional_status,) in list(
+            self._transition_door_status.items()
+        ):
+            if (
+                datetime.utcnow() - cached_transitional_status.activated
+                >= self._transition_status_timeout
+            ):
+                del self._transition_door_status[cached_door_id]
+
+        # For each door, determine the status.
+        result: Final[Dict[int, AllDoorStatus]] = {}
+        for door in doors:
+            transitional_status = self._transition_door_status.get(door.door_id)
+            if (
+                use_transitional_status
+                and transitional_status
+                and transitional_status.target_status != door.status
+            ):
+                result[door.door_id] = transitional_status.transition_status
+            else:
+                result[door.door_id] = door.status
+
+        return result
+
+    async def async_get_door_statuses(
+        self, use_transitional_status: bool = True
+    ) -> Dict[int, AllDoorStatus]:
+        """Get configured door statuses."""
+        return self._get_door_statuses(
+            await self.async_info(), use_transitional_status=use_transitional_status
+        )
 
 
 class ISmartGateApi(
     AbstractGateApi[
-        ISmartGateApiCipher, ISmartGateInfoResponse, GogoGate2ActivateResponse
+        ISmartGateApiCipher, ISmartGateInfoResponse, ISmartGateActivateResponse
     ]
 ):
     """API for interacting with iSmartGate devices."""
 
-    def __init__(self, host: str, username: str, password: str) -> None:
+    def __init__(
+        self,
+        host: str,
+        username: str,
+        password: str,
+        request_timeout: timedelta = AbstractGateApi.DEFAULT_REQUEST_TIMEOUT,
+        transition_status_timeout: timedelta = AbstractGateApi.DEFAULT_TRANSITION_STATUS_TIMEOUT,
+    ) -> None:
         """Initialize the object."""
         super().__init__(
-            host, username, password, ISmartGateApiCipher(username, password)
+            host,
+            username,
+            password,
+            ISmartGateApiCipher(username, password),
+            request_timeout=request_timeout,
+            transition_status_timeout=transition_status_timeout,
         )
 
-    def info(self) -> ISmartGateInfoResponse:
+    async def async_info(self) -> ISmartGateInfoResponse:
         """Get info about the device and doors."""
-        return element_to_ismartgate_info_response(self._info())
+        return element_to_ismartgate_info_response(await self._async_info())
 
-    def activate(self, door_id: int) -> GogoGate2ActivateResponse:
+    async def async_activate(self, door_id: int) -> ISmartGateActivateResponse:
         """Send a command to open/close/stop the door.
 
         Devices do not have a status for opening or closing. So running
         this method during an action will stop the door. It's recommended you
         use open_door() or close_door() as those methods check the status
         before running and run if needed."""
-        return element_to_gogogate2_activate_response(self._activate(door_id))
+        return element_to_ismartgate_activate_response(
+            await self._async_activate(door_id)
+        )
 
     #  pylint: disable=no-self-use
     def _get_activate_api_code(self, info: ISmartGateInfoResponse, door_id: int) -> str:
         """Get api code for activate actions."""
-        door: Final[Optional[AbstractDoor]] = get_door_by_id(door_id, info)
+        door: Final = get_door_by_id(door_id, info)
         if not door:
-            raise Exception(f"Door with id {door_id} not found.")
+            raise InvalidDoorException(door_id)
 
         return cast(ISmartGateDoor, door).apicode
 
@@ -352,27 +454,43 @@ class ISmartGateApi(
 
 class GogoGate2Api(
     AbstractGateApi[
-        GogoGate2ApiCipher, GogoGate2InfoResponse, ISmartGateActivateResponse
+        GogoGate2ApiCipher, GogoGate2InfoResponse, GogoGate2ActivateResponse
     ]
 ):
     """API for interacting with GogoGate2 devices."""
 
-    def __init__(self, host: str, username: str, password: str) -> None:
+    def __init__(
+        self,
+        host: str,
+        username: str,
+        password: str,
+        request_timeout: timedelta = AbstractGateApi.DEFAULT_REQUEST_TIMEOUT,
+        transition_status_timeout: timedelta = AbstractGateApi.DEFAULT_TRANSITION_STATUS_TIMEOUT,
+    ) -> None:
         """Initialize the object."""
-        super().__init__(host, username, password, GogoGate2ApiCipher())
+        super().__init__(
+            host,
+            username,
+            password,
+            GogoGate2ApiCipher(),
+            request_timeout=request_timeout,
+            transition_status_timeout=transition_status_timeout,
+        )
 
-    def info(self) -> GogoGate2InfoResponse:
+    async def async_info(self) -> GogoGate2InfoResponse:
         """Get info about the device and doors."""
-        return element_to_gogogate2_info_response(self._info())
+        return element_to_gogogate2_info_response(await self._async_info())
 
-    def activate(self, door_id: int) -> ISmartGateActivateResponse:
+    async def async_activate(self, door_id: int) -> GogoGate2ActivateResponse:
         """Send a command to open/close/stop the door.
 
         Devices do not have a status for opening or closing. So running
         this method during an action will stop the door. It's recommended you
         use open_door() or close_door() as those methods check the status
         before running and run if needed."""
-        return element_to_ismartgate_activate_response(self._activate(door_id))
+        return element_to_gogogate2_activate_response(
+            await self._async_activate(door_id)
+        )
 
     @staticmethod
     def _get_exception_map() -> Dict[int, ExceptionGenerator]:
